import { ConcurrentMapper } from './concurrency.js';
export function prefetchGen(iterable) {
    const it = iterable[Symbol.asyncIterator]();
    let prom = it.next();
    return {
        [Symbol.asyncIterator]() {
            return {
                next() {
                    const oldProm = prom;
                    prom = it.next();
                    return oldProm;
                },
            };
        },
    };
}
export async function* appendGen(iterA, iterB) {
    yield* iterA;
    yield* iterB;
}
export async function* prependGen(iterA, iterB) {
    yield* iterB;
    yield* iterA;
}
export async function* dropGen(iter, num) {
    let rem = num;
    for await (const elem of iter) {
        if (rem-- <= 0) {
            yield elem;
        }
    }
}
export async function* takeGen(iter, num) {
    let rem = num;
    for await (const elem of iter) {
        if (rem-- <= 0) {
            return;
        }
        yield elem;
    }
}
export async function* dropLastGen(iter, num) {
    const buf = Array(num);
    let pos = 0;
    let size = 0;
    for await (const elem of iter) {
        if (size >= num) {
            yield num ? buf[pos] : elem;
        }
        else {
            size++;
        }
        if (num) {
            buf[pos] = elem;
            pos = ((pos + 1) % num);
        }
    }
}
export async function* takeLastGen(iter, num) {
    const buf = Array(num);
    let pos = 0;
    let size = 0;
    for await (const elem of iter) {
        size = size >= num ? num : size + 1;
        if (num) {
            buf[pos] = elem;
            pos = ((pos + 1) % num);
        }
    }
    pos = size >= num ? pos : 0;
    for (let i = 0; i < size; i++) {
        yield buf[pos];
        pos = ((pos + 1) % num);
    }
}
export async function* dropWhileGen(iter, func) {
    let idx = 0;
    let yielding = false;
    for await (const elem of iter) {
        if (yielding || !(await func(elem, idx++))) {
            yielding = true;
            yield elem;
        }
    }
}
export async function* takeWhileGen(iter, func) {
    let idx = 0;
    for await (const elem of iter) {
        if (!(await func(elem, idx++))) {
            return;
        }
        yield elem;
    }
}
async function* sliceNegPosGen(iter, start, end) {
    const buf = Array(start);
    let pos = 0;
    let size = 0;
    let num = end;
    for await (const elem of iter) {
        if (size >= start && end) {
            num--;
        }
        size = size >= start ? start : size + 1;
        if (start) {
            buf[pos] = elem;
            pos = ((pos + 1) % start);
        }
    }
    pos = size >= start ? pos : 0;
    for (let i = 0; i < Math.min(size, num); i++) {
        yield buf[pos];
        pos = ((pos + 1) % start);
    }
}
export function sliceGen(iter, start, end) {
    if (start >= 0 && end == null) {
        return dropGen(iter, start);
    }
    else if (end == null) { // && start < 0
        return takeLastGen(iter, -start);
    }
    else if (start >= 0 && end >= 0) {
        return takeGen(dropGen(iter, start), end - start);
    }
    else if (start < 0 && end >= 0) {
        return sliceNegPosGen(iter, -start, end);
    }
    else if (start >= 0) { // && end < 0
        return dropLastGen(dropGen(iter, start), -end);
    }
    else { // start < 0 && end < 0
        return dropLastGen(takeLastGen(iter, -start), -end);
    }
}
export async function* filterGen(iter, func, options) {
    const concIter = new ConcurrentMapper(iter, func, options);
    for await (const elem of concIter) {
        if (elem.mapped) {
            yield elem.original;
        }
    }
}
export async function* mapGen(iter, func, options) {
    const concIter = new ConcurrentMapper(iter, func, options);
    for await (const elem of concIter) {
        yield elem.mapped;
    }
}
export async function* tapGen(iter, func, options) {
    const concIter = new ConcurrentMapper(iter, func, options);
    for await (const elem of concIter) {
        yield elem.original;
    }
}
export async function* flattenGen(iter) {
    for await (const elem of iter) {
        yield* elem;
    }
}
export async function* chunkGen(iter, num) {
    let chunk = [];
    for await (const elem of iter) {
        chunk.push(elem);
        if (chunk.length === num) {
            yield chunk;
            chunk = [];
        }
    }
    if (chunk.length) {
        yield chunk;
    }
}
export async function* chunkWhileGen(iter, func) {
    let chunk = [];
    for await (const elem of iter) {
        if (chunk.length === 0 || await func(elem, chunk[chunk.length - 1], chunk[0])) {
            chunk.push(elem);
        }
        else {
            yield chunk;
            chunk = [elem];
        }
    }
    if (chunk.length) {
        yield chunk;
    }
}
export async function* groupByGen(iter, func, options) {
    const groups = new Map();
    const concIter = new ConcurrentMapper(iter, func, options);
    for await (const elem of concIter) {
        const key = elem.mapped;
        const group = groups.get(key) ?? [];
        group.push(elem.original);
        groups.set(key, group);
    }
    yield* groups.entries();
}
export async function* uniqueGen(iter, func, options) {
    const seen = new Set();
    const concIter = new ConcurrentMapper(iter, func, options);
    for await (const elem of concIter) {
        const key = elem.mapped;
        if (!seen.has(key)) {
            yield elem.original;
            seen.add(key);
        }
    }
}
export async function* reverseGen(iter) {
    const buf = [];
    for await (const elem of iter) {
        buf.push(elem);
    }
    for (let i = buf.length - 1; i >= 0; i--) {
        yield buf[i];
    }
}
export async function* sortGen(iter, func) {
    const buf = [];
    for await (const elem of iter) {
        buf.push(elem);
    }
    for await (const elem of buf.sort(func)) {
        yield elem;
    }
}
//# sourceMappingURL=generators.js.map