import AsyncIterableBuilder from './builder.js';
import PolyAsyncIterable from './async/poly-iterable.js';
import { PolySyncIterable, IterablePolySyncIterable } from './sync/main.js';
import { isAsyncIterable, isSyncIterable } from './utils.js';
function* rangeUp(from, to, step) {
    for (let n = from; n < to; n += step) {
        yield n;
    }
}
function* rangeDown(from, to, step) {
    for (let n = from; n > to; n -= step) {
        yield n;
    }
}
/**
 * Main namespace for the creation of {@link PolySyncIterable} and {@link PolyAsyncIterable} objects.
 *
 * @public
 */
export var Poly;
(function (Poly) {
    /**
     * Create a new {@link PolySyncIterable} object from an iterable or a function that returns iterables (such as a
     * generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolySyncIterable}.
     *
     * @typeParam T - The type of the resulting iterable
     * @param iterableOrFactory - an iterable or iterable factory used to create the resulting iterable
     * @returns A {@link PolySyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function syncFrom(iterableOrFactory) {
        if (iterableOrFactory instanceof PolySyncIterable) {
            return iterableOrFactory;
        }
        const iterable = (typeof iterableOrFactory === 'function') ? iterableOrFactory() : iterableOrFactory;
        if (isSyncIterable(iterable)) {
            return new IterablePolySyncIterable(iterable);
        }
        throw Error('argument is not iterable');
    }
    Poly.syncFrom = syncFrom;
    /**
     * Create a new {@link PolyAsyncIterable} object from an iterable, async iterable, or a function that returns
     * iterables (such as a generator function or an async generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolyAsyncIterable}.
     *
     * @typeParam T - The type of the resulting async iterable
     * @param iterableOrFactory - an async iterable or async iterable factory used to create the resulting async iterable
     * @returns A {@link PolyAsyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function asyncFrom(iterableOrFactory) {
        if (iterableOrFactory instanceof PolyAsyncIterable) {
            return iterableOrFactory;
        }
        const iterable = (typeof iterableOrFactory === 'function') ? iterableOrFactory() : iterableOrFactory;
        if (isAsyncIterable(iterable)) {
            return new PolyAsyncIterable(iterable);
        }
        if (isSyncIterable(iterable)) {
            return new IterablePolySyncIterable(iterable).async();
        }
        throw Error('argument is not sync or async iterable');
    }
    Poly.asyncFrom = asyncFrom;
    function from(iterableOrFactory) {
        if (iterableOrFactory instanceof PolySyncIterable || iterableOrFactory instanceof PolyAsyncIterable) {
            return iterableOrFactory;
        }
        const iterable = (typeof iterableOrFactory === 'function') ? iterableOrFactory() : iterableOrFactory;
        if (isAsyncIterable(iterable)) {
            return new PolyAsyncIterable(iterable);
        }
        if (isSyncIterable(iterable)) {
            return new IterablePolySyncIterable(iterable);
        }
        throw Error('argument is not sync or async iterable');
    }
    Poly.from = from;
    /**
     * Return a new {@link AsyncIterableBuilder}, an iterable object that can be constructed by calling its methods.
     *
     * @remarks
     * This method for creating iterables should only be used as last resort if other functions are not enough.
     * In particular, before using this method, see if {@link Poly.buildWith} works for you.
     *
     * @typeParam T - The type of the resulting builder
     *
     * @public
     */
    function builder() {
        return new AsyncIterableBuilder();
    }
    Poly.builder = builder;
    /**
     * Create a new {@link PolyAsyncIterable} by passing the `func` function an {@link AsyncIterableBuilder} object.
     *
     * @remarks
     * This method for creating iterables should only be used as a last resort if other functions are not enough.
     *
     * @typeParam T - The type of the resulting async iterable
     * @param func - A function that will receive an object containing the methods `value`, `error` and `done`.
     *
     * @public
     */
    function buildWith(func) {
        const bld = builder();
        func(bld);
        return Poly.asyncFrom(bld);
    }
    Poly.buildWith = buildWith;
    /**
     * Return a {@link PolySyncIterable} that yields no elements upon iteration.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields no elements
     *
     * @public
     */
    function empty() {
        return syncFrom([]);
    }
    Poly.empty = empty;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.entries` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the entries of the passed object
     *
     * @public
     */
    function entries(obj) {
        return syncFrom(function* () {
            for (const key in obj) {
                if (Object.prototype.hasOwnProperty.call(obj, key)) {
                    yield [key, obj[key]];
                }
            }
        });
    }
    Poly.entries = entries;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.keys` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the keys of the passed object
     *
     * @public
     */
    function keys(obj) {
        return syncFrom(function* () {
            for (const key in obj) {
                if (Object.prototype.hasOwnProperty.call(obj, key)) {
                    yield key;
                }
            }
        });
    }
    Poly.keys = keys;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.values` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the values of the passed object
     *
     * @public
     */
    function values(obj) {
        return syncFrom(function* () {
            for (const key in obj) {
                if (Object.prototype.hasOwnProperty.call(obj, key)) {
                    yield obj[key];
                }
            }
        });
    }
    Poly.values = values;
    function range(fromOrTo, maybeTo, step = 1) {
        const fromNum = (maybeTo === undefined) ? 0 : fromOrTo;
        const toNum = (maybeTo === undefined) ? fromOrTo : maybeTo;
        if (step === 0) {
            throw new Error('"step" can\'t be 0');
        }
        const rangeFunc = (step > 0 ? rangeUp : rangeDown);
        return syncFrom(rangeFunc(fromNum, toNum, Math.abs(step)));
    }
    Poly.range = range;
    /**
     * Returns a {@link PolySyncIterable} that yields the passed argument forever, without ending.
     *
     * @typeParam T - The type of the interable
     * @param value - The reapeated element
     * @returns an infinite iterable over the same element
     *
     * @public
     */
    function repeat(value) {
        return syncFrom(function* () {
            while (true) {
                yield value;
            }
        });
    }
    Poly.repeat = repeat;
    function syncIterate(func, initValue) {
        return syncFrom(function* () {
            let nextValue = initValue;
            while (true) {
                nextValue = func(nextValue);
                yield nextValue;
            }
        });
    }
    Poly.syncIterate = syncIterate;
    function asyncIterate(func, initValue) {
        return asyncFrom(async function* () {
            let nextValue = initValue;
            while (true) {
                nextValue = await func(nextValue);
                yield nextValue;
            }
        });
    }
    Poly.asyncIterate = asyncIterate;
})(Poly || (Poly = {}));
//# sourceMappingURL=poly.js.map