import { appendGen, chunkGen, chunkWhileGen, dropGen, dropLastGen, dropWhileGen, flattenGen, groupByGen, prependGen, reverseGen, sliceGen, sortGen, takeGen, takeLastGen, takeWhileGen, tapGen, uniqueGen, } from './generators.js';
import { comparator, asserts, identity, isNotNullish } from '../utils.js';
import PolyAsyncIterable from '../async/poly-iterable.js';
/** @internal */
export const baseImpls = {};
/**
 * A `SyncIterable<T>` with a suite of methods for transforming the iteration into other iterations or to get a single
 * result from it.
 *
 * The methods of this class are intended to resemble those of `Array`, with added utilities where appropriate and made
 * for any kind of iterable.
 *
 * @public
 */
export default class PolySyncIterable {
    /**
     * Return an async version of this same iteration.
     *
     * @returns A {@link PolyAsyncIterable} that yields the same elements as `this`
     */
    async() {
        const syncIterable = this;
        const asyncIterable = {
            async *[Symbol.asyncIterator]() {
                yield* syncIterable;
            },
        };
        return new PolyAsyncIterable(asyncIterable);
    }
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolySyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    append(other) {
        asserts.isSyncIterable(other);
        return new IterablePolySyncIterable(appendGen(this, other));
    }
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * This method is an alias for {@link PolySyncIterable.append}.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolySyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    concat(other) {
        return this.append(other);
    }
    /**
     * Return a new iteration that will iterate over `other`, then over `this`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be prepended
     * @param other - Iterable to be prepended
     * @returns a new {@link PolySyncIterable} that yields the elements of `other` and then the elements of `this`
     */
    prepend(other) {
        asserts.isSyncIterable(other);
        return new IterablePolySyncIterable(prependGen(this, other));
    }
    /**
     * Return a new iteration that skips the first `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, except for the first
     * `num` elements
     */
    drop(num = 0) {
        asserts.isNonNegativeInteger(num);
        return new IterablePolySyncIterable(dropGen(this, num));
    }
    /**
     * Return a new iteration that iterates only over the first `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolySyncIterable} that yields the first `num` elements elements of `this`
     */
    take(num = 0) {
        asserts.isNonNegativeInteger(num);
        return new IterablePolySyncIterable(takeGen(this, num));
    }
    /**
     * Return a new iteration that skips the last `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to skip those if the iteration ends,
     * and so elements effectively get delayed by `num` elements.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, except for the last
     * `num` elements
     */
    dropLast(num = 0) {
        asserts.isNonNegativeInteger(num);
        return new IterablePolySyncIterable(dropLastGen(this, num));
    }
    /**
     * Return a new iteration that iterates only over the last `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to know which elements to keep.
     * and so elements effectively get delayed until the iteration ends.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolySyncIterable} that yields the last `num` elements elements of `this`
     */
    takeLast(num = 0) {
        asserts.isNonNegativeInteger(num);
        return new IterablePolySyncIterable(takeLastGen(this, num));
    }
    /**
     * Return a new iteration that skips the first few elements for which `func(element)` returns `true`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, excepts the first few for
     * which`func(element)` returns `true`
     */
    dropWhile(func) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(dropWhileGen(this, func));
    }
    takeWhile(func) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(takeWhileGen(this, func));
    }
    /**
     * Return a new iteration that starts from the `start`th element (included)
     * and ends at the `end`th element (excluded) of `this`.
     *
     * @remarks
     * Both `start` and `end` allow for negative values, in which case they refer to the nth-to-last element,
     * with n being the absolute value of the argument.  `end` might also be `undefined`, in which case the iteration is
     * not shortened on the end side, yielding up to the end, including the last element.
     * This mimics the behaviour of {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/slice | Array.slice}.
     *
     * This function will likely need a buffer, effectively delaying the yielding of elements for a while.
     *
     * @param start - The index of the first element returned
     * @param end - The index of the first element *not* returned, inclusive
     * @returns a new {@link PolySyncIterable} that yields the elements going that starts from the `start`th element
     * (included) and ends at the `end`th element (excluded) of `this`
     */
    slice(start, end) {
        asserts.isInteger(start, 'start');
        asserts.isInteger(end ?? 0, 'end');
        return new IterablePolySyncIterable(sliceGen(this, start, end));
    }
    filter(func) {
        asserts.isFunction(func);
        return baseImpls.filter(this, func);
    }
    /**
     * Return an iteration of all the elements as `this` that aren't `null` or `undefined`.
     *
     * @remarks
     * This function is a shortcut to calling {@link PolySyncIterable.filter.(:FILTER_TYPEPRED)} with a type predicate
     * function that correctly filters out `null` and `undefined` values from the iteration.  Note that other falsy values
     * will remain in the iteration, and that the return value is correctly typed to exclude `null` and `undefined`.
     *
     * @returns A new {@link PolySyncIterable} that yields the same elements as `this`
     * except for `null` or `undefined` values
     */
    filterNotNullish() {
        return this.filter(isNotNullish);
    }
    /**
     * Return an iteration of the result of calling `func(element)` for every element in `this`.
     *
     * @typeParam U - The return type of `func` and the generic type of the resulting iterable
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this`
     */
    map(func) {
        asserts.isFunction(func);
        return baseImpls.map(this, func);
    }
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the first element of the pair (the *key*) and preserving the second (the *value*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K1 - The key type of the original iterable, as passed to `func`
     * @typeParam K2 - The key type of the resulting iterable, as returned from `func`
     * @typeParam V - The value type of the original and resulting iterables, as passed to `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)` for every element of
     * `this` and uses it to replace the keys
     */
    mapKeys(func) {
        asserts.isFunction(func);
        return this.map(([k, v], index) => [func([k, v], index), v]);
    }
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the second element of the pair (the *value*) and preserving the first (the *key*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K - The key type of the original and resulting iterables, as passed to `func`
     * @typeParam V1 - The value type of the original iterable, as passed to `func`
     * @typeParam V2 - The value type of the resulting iterable, as returned from `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this` and using it to replace the values
     */
    mapValues(func) {
        asserts.isFunction(func);
        return this.map(([k, v], index) => [k, func([k, v], index)]);
    }
    /**
     * Return an iteration of the same elements as `this` after calling `func(element)` for all elements.
     *
     * @typeParam U - The return type of `func`
     * @param func - A function called for all elements
     * @returns A new {@link PolySyncIterable} that yields the same elements as `this`
     */
    tap(func) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(tapGen(this, func));
    }
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolySyncIterable} that will yield the elements of all sub-iterables
     */
    flatten() {
        return new IterablePolySyncIterable(flattenGen(this));
    }
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @remarks
     * This method is an alias of {@link PolySyncIterable.flatten}.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolySyncIterable} that will yield the elements of all sub-iterables
     */
    flat() {
        return this.flatten();
    }
    /**
     * Return an iteration of elements of the sub-iterables that result from calling `func(element)`
     * for every element in `this`.
     *
     * @remarks
     * This method is equivalent to calling {@link PolySyncIterable.map | map(func)}
     * and then {@link PolySyncIterable.flatten | flatten()}
     *
     * @typeParam U - The type of the sub-iterables returned by `func`
     * @param func - A function that takes an element of `this` and returns an iterable
     * @returns A new {@link PolySyncIterable} that yields the elements of the subiterables that results from
     * calling `func(element)` for every element of `this`
     */
    flatMap(func) {
        return this.map(func).flatten();
    }
    /**
     * Return an iteration of arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this` iteration.
     *
     * @remarks
     * All chunks except possibly the last one will have exactly `num` elements.  The last chunk will have less elements
     * if the number of elements in the iteration is not divisible by `num`.  No chunk will ever be returned empty or
     * have more than `num` elements.
     *
     * @param num - Size of the chunks
     * @returns A new {@link PolySyncIterable} that yields arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this`
     */
    chunk(num = 1) {
        asserts.isPositiveInteger(num);
        return new IterablePolySyncIterable(chunkGen(this, num));
    }
    /**
     * Return an iteration of arrays with elements of this separated based on the result of calling `func(elements)`.
     *
     * @remarks
     * The chunking process works by keeping an open _current chunk_ and calling `func` to decide whether the next
     * element of the iteration will be part of the _current chunk_ or if it will be part of a new chunk.
     *
     * To do this, `func` will be called for every element of the iteration except the first, which will automatically
     * become part of the first chunk.  If `func` returns `true`, the element will be part of the current chunk, and
     * if it returns `false`, the _current chunk_ is closed and the element becomes the first element if the new
     * _current chunk_.
     * The arguments passed to `func` will be, in order:
     *   - `elem` - The element being currently processed
     *   - `lastElem` - The last element that was added to the current chunk
     *   - `firstElem` - The first element of the current chunk (might be the same as `lastElem`)
     *
     * All elements will be part of a chunk, and no chunk will ever be empty.
     *
     * @param func - A function that decides if an element is part of the current chunk or initiates a new one
     * @returns A new {@link PolySyncIterable} that yields arrays with the elements of `this` as separated by `func`
     */
    chunkWhile(func) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(chunkWhileGen(this, func));
    }
    /**
     * Return an iteration of group pairs, where the first element is a _group key_ and the second is an iterable of all
     * the elements for which `func(element)` returned the key.
     *
     * @remarks
     * This method is intended to be combined with {@link PolySyncIterable.toObject | toObject}
     * or {@link PolySyncIterable.toMap | toMap}, thus behaving like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupBy | Array.groupBy} and
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupByToMap | Array.groupByToMap}
     * respectively, but without losing the ablity to further process the iteration, such as by mapping, filtering, etc.
     *
     * @typeParam K - Type of the keys used to group elements
     * @param func - A function that returns the grouping key of each element
     * @returns A new {@link PolySyncIterable} of group pairs with the key and the group
     */
    groupBy(func) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(groupByGen(this, func));
    }
    /**
     * Return an iteration of unique elements, where two elements are considered equal if the result of `func(element)` is
     * the same for both elements.
     *
     * @remarks
     * Note that the first element seen with a specific key is always the one yielded, and every other element afterwards
     * is ignored.
     *
     * If no key-mapping function is given, the elements theselves are used as keys.
     * This is likely _not_ what you want in most situations unless elements are primitive types.
     *
     * @param func - A function that returns a _key_ used for uniqueness checks.
     * If not passed, an identitity function is used.
     * @returns A new {@link PolySyncIterable} only elements for which `func(element)` returns a value that hasn't
     * been seen before
     */
    unique(func = identity) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(uniqueGen(this, func));
    }
    /**
     * Return an iteration of the elements of `this` in reverse order.
     *
     * @remarks
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @returns A new {@link PolySyncIterable} that yields the elements of `this` in reverse order
     */
    reverse() {
        return new IterablePolySyncIterable(reverseGen(this));
    }
    /**
     * Return an iteration of the elements of `this` sorted according to `func`
     *
     * @remarks
     * The sort function `func` is used to call
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * on an array of all the elements.
     * However, the default comparator function will sort elements according to the `<` and `>` operators defined on
     * their own type, of always sorting lexicagraphically like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * does.
     *
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @param func - A comparator function
     * @returns A new {@link PolySyncIterable} that yields the elements of `this` sorted according to `func`
     */
    sort(func = comparator) {
        asserts.isFunction(func);
        return new IterablePolySyncIterable(sortGen(this, func));
    }
    /**
     * Return an array of all elements of this iteration in the same order that were yielded.
     *
     * @returns An array that contains the same elements as this iteration, in the same order
     */
    toArray() {
        return Array.from(this);
    }
    toPartitionArrays(func) {
        const trues = [];
        const falses = [];
        let idx = 0;
        for (const elem of this) {
            if (func(elem, idx++)) {
                trues.push(elem);
            }
            else {
                falses.push(elem);
            }
        }
        return [trues, falses];
    }
    /**
     * Return an object made from the entries of `this`.
     * This method is roughly equivalent to calling `Object.fromEntires(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs where the first component is a valid object key type.
     *
     * @returns An object composed of the entries yielded by this iterable.
     */
    toObject() {
        const object = {};
        for (const [key, value] of this) {
            object[key] = value;
        }
        return object;
    }
    /**
     * Return a `Map` made from the entries of `this`.
     * This method is roughly equivalent to calling `new Map(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @returns A `Map` composed of the entries yielded by this iterable.
     */
    toMap() {
        const map = new Map();
        for (const [key, value] of this) {
            map.set(key, value);
        }
        return map;
    }
    find(func) {
        asserts.isFunction(func);
        let idx = 0;
        for (const elem of this) {
            if (func(elem, idx++)) {
                return elem;
            }
        }
        return undefined;
    }
    findLast(func) {
        asserts.isFunction(func);
        let found;
        let idx = 0;
        for (const elem of this) {
            if (func(elem, idx++)) {
                found = elem;
            }
        }
        return found;
    }
    /**
     * Returns the index of the first element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The index of the first element of the iteration for which `func` returned `true`
     */
    findIndex(func) {
        asserts.isFunction(func);
        let idx = 0;
        for (const elem of this) {
            if (func(elem, idx)) {
                return idx;
            }
            idx++;
        }
        return -1;
    }
    /**
     * Returns the index of the last element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The index of the last element of the iteration for which `func` returned `true`
     */
    findLastIndex(func) {
        asserts.isFunction(func);
        let foundIndex = -1;
        let idx = 0;
        for (const elem of this) {
            if (func(elem, idx)) {
                foundIndex = idx;
            }
            idx++;
        }
        return foundIndex;
    }
    /**
     * Returns whether an element is present in this iteration.
     *
     * @remarks
     * If the element is found in the iteration, no more elements are iterated.
     *
     * @param obj - The element to search in the iteration
     * @returns Whether `obj` is present in this iteration or not
     */
    includes(obj) {
        for (const elem of this) {
            if (Object.is(obj, elem) || obj === elem) {
                return true;
            }
        }
        return false;
    }
    /**
     * Returns `true` if calling `func(element)` returns `true` for at least one element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `true`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @returns Whether calling `func` returned `true` on at least one element.
     */
    some(func) {
        asserts.isFunction(func);
        let idx = 0;
        for (const item of this) {
            if (func(item, idx++)) {
                return true;
            }
        }
        return false;
    }
    /**
     * Returns `true` if calling `func(element)` returns `true` for every element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `false`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @returns Whether calling `func` returned `true` for all elements.
     */
    every(func) {
        asserts.isFunction(func);
        let idx = 0;
        for (const item of this) {
            if (!func(item, idx++)) {
                return false;
            }
        }
        return true;
    }
    reduce(reducer, init) {
        asserts.isFunction(reducer);
        let accumulated = init;
        let isFirst = (accumulated === undefined);
        let idx = 0;
        for (const elem of this) {
            accumulated = isFirst ? elem : reducer(accumulated, elem, idx++);
            isFirst = false;
        }
        return accumulated;
    }
    /**
     * Return the number of elements on this iteration.
     *
     * @returns The number of elements in this iteration
     */
    count() {
        let count = 0;
        for (const _ of this) {
            count++;
        }
        return count;
    }
    /**
     * Call a function for each element of `this` iteration.
     *
     * @param func - A function to be called for every element of the iteration
     */
    forEach(func) {
        asserts.isFunction(func);
        let idx = 0;
        for (const elem of this) {
            func(elem, idx++);
        }
    }
    /**
     * Return the result of joining the elements of `this` with the given `glue`, or `','` if no glue is given.
     *
     * @remarks
     * `null` or `undefined` elements are treated as empty strings.
     *
     * @param glue - The string to use for joining the elements
     * @returns A string concatenating all elements of `this` using the given `glue`
     */
    join(glue = ',') {
        let str = '';
        let first = true;
        for (const elem of this) {
            str += (first ? '' : glue) + (elem == null ? '' : elem);
            first = false;
        }
        return str;
    }
    /**
     * Perform this iteration doing nothing.
     */
    complete() {
        /* eslint-disable-next-line no-unused-vars */
        for (const elem of this) {
            /* do nothing, just iterate */
        }
    }
    /**
     * Returns a tuple containing `num` iterables that will yield independent
     * copies of the elements yielded by `this`.
     *
     * @remarks
     * Note that, as with every other method of this class, this instance is unusable after calling this method.
     *
     * In order to provide a truly independent iteration for all returned iterables, a buffer is kept, which can grow as
     * big as the whole iteration in certain circumstances.  The buffer is filled as fast as the fastest iterable requests
     * new items, and emptied as fast as the slowest iterable requests new items.
     *
     * Note that for synchronous iterations, it's common to end up with a full buffer if the returned duplicated elements
     * are used in sequence.  In this situation, it might be more useful to simply convert the iteration to an array and
     * pass it around, rather than pay the overhead of this method.
     *
     * @param num - the number of copies to be returned
     * @returns An array of `num` elements containing independent copies of this iterable
     */
    duplicate(num) {
        asserts.isNonNegativeInteger(num);
        const it = this[Symbol.iterator]();
        let offsIndex = 0; // index offset for when the buffer is shrunk
        const buffer = [];
        function* generator() {
            let currIndex = 0;
            while (true) {
                const index = currIndex - offsIndex;
                if (buffer[index] == null) {
                    const item = it.next();
                    buffer[index] = { item, pending: num };
                }
                if (buffer[index].item.done) {
                    return;
                }
                currIndex += 1;
                buffer[index].pending -= 1;
                yield buffer[index].item.value;
                while (buffer.length > 0 && buffer[0].pending <= 0 && !buffer[0].item.done) {
                    buffer.shift();
                    offsIndex += 1;
                }
            }
        }
        return Array(num).fill(null)
            .map(() => new IterablePolySyncIterable(generator()));
    }
}
/**
 * @internal
 *
 * A {@link PolySyncIterable} implementation that iterates over the object passed as its constructor argument.
 * This class in intended as a default implementation when specific, more efficient implementation are not available.
 */
export class IterablePolySyncIterable extends PolySyncIterable {
    #iterable;
    /** @internal */
    constructor(iterable) {
        super();
        this.#iterable = iterable;
    }
    [Symbol.iterator]() {
        return this.#iterable[Symbol.iterator]();
    }
}
//# sourceMappingURL=poly-sync-iterable.js.map