/**
 * A library for working with iterables and async iterables in a functional style, by transforming the iterations on
 * the fly and getting results from them.
 *
 * @remarks
 * The entry point for this library is the {@link Poly} namespace, and the main iterable classes are
 * {@link PolySyncIterable} and {@link PolyAsyncIterable}.
 *
 * @packageDocumentation
 */

/**
 * A function that receives an element (`elem`) of the iteration and the first and last element of a chunk, and returns
 * a boolean value or a `Promise` to a boolean value representing whether `elem` should be added to the current chunk
 * (if `true`) or be the start of a new chunk (if `false`)
 *
 * @typeParam T - The type of the iteration elements
 * @param elem - An element of the iteration
 * @param lastElem - The last element of the current chunk
 * @param firstElem - The first element of the current chunk
 * @returns A boolean value or a `Promise` to a boolean value representing whether `elem` should be added to the
 * current chunk (`true`) or be the start of a new chunk (`false`)
 *
 * @public
 */
export declare type AsyncChunkingPredicate<T> = (elem: T, lastElem: T, firstElem: T) => boolean | PromiseLike<boolean>;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and returns a different object or a
 * `Promise` to a different object
 *
 * @typeParam T - The type of the `elem` argument
 * @typeParam U - The return type of the function
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns Some other value or a `Promise` to some other value
 *
 * @public
 */
export declare type AsyncIndexedMapping<T, U> = (elem: T, index: number) => U | PromiseLike<U>;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and returns a `boolean` value or a
 * `Promise` to a `boolean` value.
 *
 * @typeParam T - The type of the `elem` argument
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns A `boolean` value or a `Promise` to a `boolean` value
 *
 * @public
 */
export declare type AsyncIndexedPredicate<T> = (elem: T, index: number) => boolean | PromiseLike<boolean>;

/**
 * A function that receives an accumulated result, an element of an iteration, and returns a new accumulated result or
 * a promise to a new accumulated result for the next call or as a final return value.
 *
 * @typeParam T - The type of the iteration elements
 * @typeParam U - The type of the accumulated result
 * @param acc - The previously accumulated value
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns The new accumulated result or a promise to the new accumulated result
 *
 * @public
 */
export declare type AsyncIndexedReducer<T, U> = (acc: U, item: T, index: number) => U | PromiseLike<U>;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and either returns a `Promise` to
 * nothing or doesn't return anything
 *
 * @typeParam T - The type of the `elem` argument
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns a `Promise` to an ignored value or nothing
 *
 * @public
 */
export declare type AsyncIndexedRunnable<T> = (elem: T, index: number) => void | PromiseLike<void>;

/**
 * A class that helps with building an
 * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Symbol/asyncIterator | AsyncIterable}
 * from a non-structured source.
 *
 * In order to create the iteration, you must call the {@link AsyncIterableBuilder.value | value},
 * {@link AsyncIterableBuilder.error | error} and {@link AsyncIterableBuilder.done | done} methods with
 * appropriate arguments.
 *
 * @public
 */
export declare class AsyncIterableBuilder<T> implements AsyncIterable<T> {
    private _currentPromiseFuncs?;
    private _pendingValues;
    private _pendingError?;
    private _pendingDone;
    /** @internal */
    constructor();
    /**
     * Gives the underlying iterable a new value to be yielded.
     *
     * @remarks
     * Objects given for iteration will be buffered until they are requested, and are guaranteed to be yielded before
     * errors and before finishing the iteration.
     *
     * If this method is called after {@link AsyncIterableBuilder.error | error}
     * or {@link AsyncIterableBuilder.done | done},
     * the given object is ignored.
     *
     * @param obj - The object to be yielded by the underlying iterable
     */
    value(obj: T): void;
    /**
     * Makes the underlying iterable throw the given `error` object.
     *
     * @remarks
     * Any values that were given with {@link AsyncIterableBuilder.value | value} will be yielded before the error is
     * thrown.
     *
     * Calling {@link AsyncIterableBuilder.value | value} or {@link AsyncIterableBuilder.done | done} after calling this
     * method or calling this method after calling {@link AsyncIterableBuilder.done | done} will act as a no-op.
     *
     * @param error - The error to be thrown by the underlying iterable
     */
    error(error: Error): void;
    /**
     * Makes the underlying iterable finish the iteration.
     *
     * @remarks
     * Any values that were given with {@link AsyncIterableBuilder.value | value} will be yielded
     * before the iteration ends.
     *
     * Calling {@link AsyncIterableBuilder.value | value} or {@link AsyncIterableBuilder.error | error} after calling this
     * method or calling this method after calling {@link AsyncIterableBuilder.error | error} will act as a no-op.
     */
    done(): void;
    /**
     * Get a {@link PolyAsyncIterable} that iterates the elements as determined by calls to `this` object's
     * {@link AsyncIterableBuilder.value | value}, {@link AsyncIterableBuilder.error | error} and
     * {@link AsyncIterableBuilder.done | done} methods.
     *
     * @returns A {@link PolyAsyncIterable} that yields elements as determined by the calls to `this`' methods
     */
    toPolyAsyncIterable(): PolyAsyncIterable<T>;
    /**
     * Allows this class to work as a regular `AsyncIterable<T>`
     *
     * @returns an async iterable that yields elements as determined by the calls to `this`' methods
     */
    [Symbol.asyncIterator](): AsyncIterator<T>;
}

/**
 * A function that receives an element (`elem`) of the iteration and the first and last element of a chunk, and returns
 * a boolean value representing whether `elem` should be added to the current chunk (if `true`) or be the start of a new
 * chunk (if `false`)
 *
 * @typeParam T - The type of the iteration elements
 * @param elem - An element of the iteration
 * @param lastElem - The last element of the current chunk
 * @param firstElem - The first element of the current chunk
 * @returns Whether `elem` should be added to the current chunk (`true`) or be the start of a new chunk (`false`)
 *
 * @public
 */
export declare type ChunkingPredicate<T> = (elem: T, lastElem: T, firstElem: T) => boolean;

/**
 * A function that receives two objects `elemA` and `elemB` and returns a number value that is negative if `elemA`
 * should be sorted before `elemB`, positive if `elemA` should be sorted after `elemB`, or `0` if they should be sorted
 * at the same position.
 *
 * @typeParam T - The type of the iteration elements
 * @param elemA - the first element
 * @param elemB - the second element
 * @returns A number representing the sorting of `elemA` and `elemB` with respect to each other, as defined above
 *
 * @public
 */
export declare type Comparator<T> = (elemA: T, elemB: T) => number;

/**
 * Options for concurrency.
 *
 * @public
 */
export declare interface ConcurrencyOptions {
    /** Maximum amount of promises to launch at the same time (defaults to 1) */
    concurrency?: number;
    /** Maximum amount of intermediate results to store (defaults to twice the concurrency value) */
    bufferSize?: number;
}

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and returns a different object
 *
 * @typeParam T - The type of the `elem` argument
 * @typeParam U - The return type of the function
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns Some other value
 *
 * @public
 */
export declare type IndexedMapping<T, U> = (elem: T, index: number) => U;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and returns a `boolean` value.
 *
 * @typeParam T - The type of the `elem` argument
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns A `boolean` value
 *
 * @public
 */
export declare type IndexedPredicate<T> = (elem: T, index: number) => boolean;

/**
 * A function that receives an accumulated result, an element of an iteration, and returns a new accumulated result
 * for the next call or as a final return value.
 *
 * @typeParam T - The type of the iteration elements
 * @typeParam U - The type of the accumulated result
 * @param acc - The previously accumulated value
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns The new accumulated result
 *
 * @public
 */
export declare type IndexedReducer<T, U> = (acc: U, item: T, index: number) => U;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and doesn't return anything
 *
 * @typeParam T - The type of the `elem` argument
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 *
 * @public
 */
export declare type IndexedRunnable<T> = (elem: T, index: number) => void;

/**
 * A function that receives an object (`elem`) and its `index` in the iteration and returns a `boolean` value indicating
 * if `elem` is of the generic type `U`
 *
 * @typeParam T - The type of the `elem` argument
 * @typeParam U - The type asserted by this function
 * @param elem - An element of the iteration
 * @param index - The index of `elem` in the iteration
 * @returns A `boolean` value indicating if `elem` is of type `U`
 *
 * @public
 */
export declare type IndexedTypePredicate<T, U extends T> = (elem: T, index: number) => elem is U;

/**
 * Main namespace for the creation of {@link PolySyncIterable} and {@link PolyAsyncIterable} objects.
 *
 * @public
 */
declare namespace Poly {
    /**
     * A function with no arguments that returns an `Iterable`
     * @typeParam T - The generic type of the returned `Iterable`
     */
    type IterableFactory<T> = () => Iterable<T>;
    /**
     * A function with no arguments that returns an `AsyncIterable`
     * @typeParam T - The generic type of the returned `Iterable`
     */
    type AsyncIterableFactory<T> = () => AsyncIterable<T>;
    /**
     * Create a new {@link PolySyncIterable} object from an iterable or a function that returns iterables (such as a
     * generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolySyncIterable}.
     *
     * @typeParam T - The type of the resulting iterable
     * @param iterableOrFactory - an iterable or iterable factory used to create the resulting iterable
     * @returns A {@link PolySyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function syncFrom<T>(iterableOrFactory: Iterable<T> | IterableFactory<T>): PolySyncIterable<T>;
    /**
     * Create a new {@link PolyAsyncIterable} object from an iterable, async iterable, or a function that returns
     * iterables (such as a generator function or an async generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolyAsyncIterable}.
     *
     * @typeParam T - The type of the resulting async iterable
     * @param iterableOrFactory - an async iterable or async iterable factory used to create the resulting async iterable
     * @returns A {@link PolyAsyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function asyncFrom<T>(iterableOrFactory: Iterable<T> | IterableFactory<T> | AsyncIterable<T> | AsyncIterableFactory<T>): PolyAsyncIterable<T>;
    /**
     * Create a new {@link PolyAsyncIterable} object from an iterable, async iterable, or a function that returns
     * iterables (such as a generator function or an async generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolyAsyncIterable}.
     *
     * @typeParam T - The type of the resulting async iterable
     * @param iterableOrFactory - an async iterable or async iterable factory used to create the resulting async iterable
     * @returns A {@link PolyAsyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function from<T>(iterableOrFactory: AsyncIterable<T> | AsyncIterableFactory<T>): PolyAsyncIterable<T>;
    /**
     * Create a new {@link PolySyncIterable} object from an iterable or a function that returns iterables (such as a
     * generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolySyncIterable}.
     *
     * @typeParam T - The type of the resulting iterable
     * @param iterableOrFactory - an iterable or iterable factory used to create the resulting iterable
     * @returns A {@link PolySyncIterable} that iterates over the same elements as the passed iterable
     *
     * @public
     */
    function from<T>(iterableOrFactory: Iterable<T> | IterableFactory<T>): PolySyncIterable<T>;
    /**
     * Create a new {@link PolySyncIterable} or {@link PolyAsyncIterable} object from a sync or async iterable or a
     * function that returns sync or async iterables (such as a sync or async generator function).
     *
     * If the passed argument is a function, it will be called with no arguments and its return value will be used to
     * create the resulting {@link PolySyncIterable} or {@link PolyAsyncIterable}.
     *
     * Whether the resulting iterable is sync or async depends on the input being sync or async.
     *
     * @typeParam T - The type of the resulting iterable
     * @param iterableOrFactory - a sync or async iterable or iterable factory used to create the resulting iterable
     * @returns A {@link PolySyncIterable} or {@link PolyAsyncIterable} that iterates over the same elements as the
     * passed iterable
     *
     * @public
     */
    function from<T>(iterableOrFactory: Iterable<T> | IterableFactory<T> | AsyncIterable<T> | AsyncIterableFactory<T>): PolySyncIterable<T> | PolyAsyncIterable<T>;
    /**
     * Return a new {@link AsyncIterableBuilder}, an iterable object that can be constructed by calling its methods.
     *
     * @remarks
     * This method for creating iterables should only be used as last resort if other functions are not enough.
     * In particular, before using this method, see if {@link Poly.buildWith} works for you.
     *
     * @typeParam T - The type of the resulting builder
     *
     * @public
     */
    function builder<T>(): AsyncIterableBuilder<T>;
    /**
     * Create a new {@link PolyAsyncIterable} by passing the `func` function an {@link AsyncIterableBuilder} object.
     *
     * @remarks
     * This method for creating iterables should only be used as a last resort if other functions are not enough.
     *
     * @typeParam T - The type of the resulting async iterable
     * @param func - A function that will receive an object containing the methods `value`, `error` and `done`.
     *
     * @public
     */
    function buildWith<T>(func: (builder: AsyncIterableBuilder<T>) => void): PolyAsyncIterable<T>;
    /**
     * Return a {@link PolySyncIterable} that yields no elements upon iteration.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields no elements
     *
     * @public
     */
    function empty<T = never>(): PolySyncIterable<T>;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.entries` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the entries of the passed object
     *
     * @public
     */
    function entries<K extends string | number | symbol, V>(obj: Record<K, V>): PolySyncIterable<[K, V]>;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.keys` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the keys of the passed object
     *
     * @public
     */
    function keys<K extends string | number | symbol>(obj: Record<K, unknown>): PolySyncIterable<K>;
    /**
     * Returns a {@link PolySyncIterable} that yields the same elements as `Object.values` would.
     *
     * @typeParam T - Type of the resulting iterable
     * @returns An iterable that yields the values of the passed object
     *
     * @public
     */
    function values<V>(obj: Record<string | number | symbol, V>): PolySyncIterable<V>;
    /**
     * Returns a {@link PolySyncIterable} that yields numbers from 0 up to, but not including, `to`, with a step of 1.
     *
     * @param to - Exclusive upper bound of the iteration
     * @returns - An iterable over the numbers from 0 to `to` with a step of 1
     *
     * @public
     */
    function range(to: number): PolySyncIterable<number>;
    /**
     * Returns a {@link PolySyncIterable} that yields numbers from `from` up to, but not including, `to`,
     * with a step of 1.
     *
     * @param from - Inclusive lower bound of the iteration
     * @param to - Exclusive upper bound of the iteration
     * @returns - An iterable over the numbers from `from` to `to` with a step of 1
     *
     * @public
     */
    function range(from: number, to: number): PolySyncIterable<number>;
    /**
     * Returns a {@link PolySyncIterable} that yields numbers from `from` up to, but not including, `to`, with a step of
     * `step`.
     *
     * @param from - Inclusive lower bound of the iteration
     * @param to - Exclusive upper bound of the iteration
     * @param step - Distance between the iterated numbers
     * @returns - An iterable over the numbers from `from` to `to` with a step of `step`
     *
     * @public
     */
    function range(from: number, to: number, step?: number): PolySyncIterable<number>;
    /**
     * Returns a {@link PolySyncIterable} that yields the passed argument forever, without ending.
     *
     * @typeParam T - The type of the interable
     * @param value - The reapeated element
     * @returns an infinite iterable over the same element
     *
     * @public
     */
    function repeat<T>(value: T): PolySyncIterable<T>;
    /**
     * Returns a {@link PolySyncIterable} that will yield the values returned from calling `func` with the value last
     * returned, or `undefined` when called for the first time.
     *
     * @remarks
     * `func` will be called initially with `undefined`.  After that, each element returned by calling it will be yielded
     * as part of the resulting iterable and then passed to next call to `func`.
     *
     * Note that there is no way of yielding a different value to that passed to the next function call, and that the
     * resulting iterable will not end.  If this is undesired, either use some of the operators on the resulting iterable
     * (such as {@link PolySyncIterable.map | map} or {@link PolySyncIterable.take | take}
     * / {@link PolySyncIterable.take | takeWhile}) or use a different approach to create the iterable.
     *
     * @typeParam T - The type of the interable and return type of `func`
     * @param func - The function that will be called to generate new elements of the iteration
     * @returns an infinite iterable that yields the return values from calling `func` repeatedly
     *
     * @public
     */
    function syncIterate<T>(func: (lastValue: T | undefined) => T): PolySyncIterable<T>;
    /**
     * Returns a {@link PolySyncIterable} that will yield the values returned from calling `func` with the value last
     * returned, or `initValue` when called for the first time.
     *
     * @remarks
     * `func` will be called initially with `initValue`.  After that, each element returned by calling it will be yielded
     * as part of the resulting iterable and then passed to next call to `func`.  Note that `initValue` will not be
     * part of the iteration.
     *
     * Note that there is no way of yielding a different value to that passed to the next function call, and that the
     * resulting iterable will not end.  If this is undesired, either use some of the operators on the resulting iterable
     * (such as {@link PolySyncIterable.map | map} or {@link PolySyncIterable.take | take}
     * / {@link PolySyncIterable.take | takeWhile}) or use a different approach to create the iterable.
     *
     * @typeParam T - The type of the interable and return type of `func`
     * @param func - The function that will be called to generate new elements of the iteration
     * @returns an infinite iterable that yields the return values from calling `func` repeatedly
     *
     * @public
     */
    function syncIterate<T>(func: (lastValue: T) => T, initValue: T): PolySyncIterable<T>;
    /**
     * Returns a {@link PolyAsyncIterable} that will yield the values returned from calling `func` with the value last
     * returned, or `undefined` when called for the first time.
     *
     * @remarks
     * `func` will be called initially with `undefined`.  After that, each element returned by calling it will be yielded
     * as part of the resulting iterable and then passed to next call to `func`.
     *
     * Note that there is no way of yielding a different value to that passed to the next function call, and that the
     * resulting iterable will not end.  If this is undesired, either use some of the operators on the resulting iterable
     * (such as {@link PolyAsyncIterable.map | map} or {@link PolyAsyncIterable.take | take}
     * / {@link PolyAsyncIterable.take | takeWhile}) or use a different approach to create the iterable.
     *
     * @typeParam T - The type of the interable and return type of `func`
     * @param func - The function that will be called to generate new elements of the iteration
     * @returns an infinite iterable that yields the return values from calling `func` repeatedly
     *
     * @public
     */
    function asyncIterate<T>(func: (lastValue: T | undefined) => T | Promise<T>): PolyAsyncIterable<T>;
    /**
     * Returns a {@link PolyAsyncIterable} that will yield the values returned from calling `func` with the value last
     * returned, or `initValue` when called for the first time.
     *
     * @remarks
     * `func` will be called initially with `initValue`.  After that, each element returned by calling it will be yielded
     * as part of the resulting iterable and then passed to next call to `func`.  Note that `initValue` will not be
     * part of the iteration.
     *
     * Note that there is no way of yielding a different value to that passed to the next function call, and that the
     * resulting iterable will not end.  If this is undesired, either use some of the operators on the resulting iterable
     * (such as {@link PolyAsyncIterable.map | map} or {@link PolyAsyncIterable.take | take}
     * / {@link PolyAsyncIterable.take | takeWhile}) or use a different approach to create the iterable.
     *
     * @typeParam T - The type of the interable and return type of `func`
     * @param func - The function that will be called to generate new elements of the iteration
     * @returns an infinite iterable that yields the return values from calling `func` repeatedly
     *
     * @public
     */
    function asyncIterate<T>(func: (lastValue: T) => T | Promise<T>, initValue: T): PolyAsyncIterable<T>;
}
export { Poly }
export default Poly;

/**
 * An `AsyncIterable<T>` with a suite of methods for transforming the iteration into other iterations or to get a
 * single result from it.
 *
 * This class works as an async version of {@link PolySyncIterable}, but all methods accept async function where
 * possible and will always return either `PolyAsyncIterables` or a `Promise` to a value.
 *
 * ## Concurrency
 * Many operations of this class accept as a final argument an {@link ConcurrencyOptions | options object} than can
 * specify some options for concurrent operations.
 *
 * - The `concurrency` option will specify how many times whatever `func` you pass is called before waiting for its
 * results.  Effectively, this is the number of promises that can be pending at the same time.  If not specified, it
 * will default to 0, meaning no concurrency.  Must be a non-negative integer.
 * - The `bufferSize` option will specify the size of the internal buffer used to store the pending and completed
 * promises.  Effectively, this is how many results will be prefetched.  If not specified, it will default to
 * `concurrency`, meaning no extra intermediate results are stored.  Must be a positive integer greater or equal to
 * `concurrency`.
 *
 * A concurrency value of 0 acts the same as a 1 concurrency-wise, but disables the concurrency completely, preventing
 * any values to be requested before actually needed.
 *
 * Specifying concurrency greater or equal to 1 will make more elements be requested of the previous iterator than
 * actually requested of the current one, similarly to {@link PolyAsyncIterable.prefetch}.
 *
 * @public
 */
export declare class PolyAsyncIterable<T> implements AsyncIterable<T> {
    #private;
    /** @internal */
    constructor(iterable: AsyncIterable<T>);
    /**
     * Allows this class to work as a regular `AsyncIterable<T>`
     *
     * @returns an async iterable that will yield the same elements as the iterable used to create this instance
     */
    [Symbol.asyncIterator](): AsyncIterator<T, any, undefined>;
    /**
     * Return `this`.
     *
     * @remarks
     * This method is present in this class for compatibility with {@link PolySyncIterable.async}, but it's never
     * necessary to call it for anything.
     *
     * @returns A {@link PolyAsyncIterable} that yields the same elements as `this`
     */
    async(): PolyAsyncIterable<T>;
    /**
     * Return the same iteration, but with its elements requested with anticipation to allow for asynchronous operations
     * to begin and reduce wait times.
     *
     * @remarks
     * When yielding an element of this iterable, the next one will be also requested internally, so that any
     * asynchronous operations are started before their results are needed.
     *
     * Note that after calling this mehtod, more elements than strictly needed might be requested from the previous
     * iterable, triggering any potential side effects.
     *
     * @returns a new {@link PolyAsyncIterable} that prefetched the iterated elementsd
     */
    prefetch(): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolyAsyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    append<U>(other: Iterable<U> | AsyncIterable<U>): PolyAsyncIterable<T | U>;
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * This method is an alias for {@link PolyAsyncIterable.append}.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolySyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    concat<U>(other: Iterable<U> | AsyncIterable<U>): PolyAsyncIterable<T | U>;
    /**
     * Return a new iteration that will iterate over `other`, then over `this`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be prepended
     * @param other - Iterable to be prepended
     * @returns a new {@link PolyAsyncIterable} that yields the elements of `other` and then the elements of `this`
     */
    prepend<U>(other: Iterable<U> | AsyncIterable<U>): PolyAsyncIterable<T | U>;
    /**
     * Return a new iteration that skips the first `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolyAsyncIterable} that yields the same the elements of `this`, except for the first
     * `num` elements
     */
    drop(num: number): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that iterates only over the first `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolyAsyncIterable} that yields the first `num` elements elements of `this`
     */
    take(num: number): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that skips the last `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to skip those if the iteration ends,
     * and so elements effectively get delayed by `num` elements.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolyAsyncIterable} that yields the same the elements of `this`, except for the last
     * `num` elements
     */
    dropLast(num: number): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that iterates only over the last `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to know which elements to keep.
     * and so elements effectively get delayed until the iteration ends.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolyAsyncIterable} that yields the last `num` elements elements of `this`
     */
    takeLast(num: number): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that skips the first few elements for which `func(element)` returns `true`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolyAsyncIterable} that yields the same the elements of `this`, excepts the first few for
     * which`func(element)` returns `true`
     */
    dropWhile(func: AsyncIndexedPredicate<T>): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that yields the first few elements for which `func(element)` returns `true`.
     *
     * @remarks
     * Because the `func` argument is a type predicate, the result iteration will have the type asserted by `func`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolyAsyncIterable} that yields the same the elements of `this` as long as `func(element)`
     * returns `true`, correctly narrowed to the type asserted by `func`
     */
    takeWhile<U extends T>(func: IndexedTypePredicate<T, U>): PolyAsyncIterable<U>;
    /**
     * Return a new iteration that yields the first few elements for which `func(element)` returns `true`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolyAsyncIterable} that yields the same the elements of `this` as long as `func(element)`
     * returns `true`
     */
    takeWhile(func: AsyncIndexedPredicate<T>): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that starts from the `start`th element (included)
     * and ends at the `end`th element (excluded) of `this`.
     *
     * @remarks
     * Both `start` and `end` allow for negative values, in which case they refer to the nth-to-last element,
     * with n being the absolute value of the argument.  `end` might also be `undefined`, in which case the iteration is
     * not shortened on the end side, yielding up to the end, including the last element.
     * This mimics the behaviour of {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/slice | Array.slice}.
     *
     * This function will likely need a buffer, effectively delaying the yielding of elements for a while.
     *
     * @param start - The index of the first element returned
     * @param end - The index of the first element *not* returned, inclusive
     * @returns a new {@link PolyAsyncIterable} that yields the elements going that starts from the `start`th element
     * (included) and ends at the `end`th element (excluded) of `this`
     */
    slice(start: number, end?: number): PolyAsyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` for which `func(element)` returns `true`.
     *
     * @remarks
     * Because the `func` argument is a type predicate, the result iteration will have the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`, if any
     * @param func - The function to be called on all elements
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} with only elements for which `func(element)` returned true, correctly
     * narrowed to the type asserted by `func`
     *
     * {@label FILTER_TYPEPRED}
     */
    filter<U extends T>(func: IndexedTypePredicate<T, U>, options?: ConcurrencyOptions): PolyAsyncIterable<U>;
    /**
     * Return an iteration of the elements of `this` for which `func(element)` returns `true`.
     *
     * @param func - The function to be called on all elements
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} with only elements for which `func(element)` returned
     */
    filter(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): PolyAsyncIterable<T>;
    /**
     * Return an iteration of all the elements as `this` that aren't `null` or `undefined`.
     *
     * @remarks
     * This function is a shortcut to calling {@link PolyAsyncIterable.filter.(:FILTER_TYPEPRED) filter} with a type
     * predicate function that correctly filters out `null` and `undefined` values from the iteration.  Note that other
     * falsy values will remain in the iteration, and that the return value is correctly typed to exclude
     * `null` and `undefined`.
     *
     * @returns A new {@link PolyAsyncIterable} that yields the same elements as `this`
     * except for `null` or `undefined` values
     */
    filterNotNullish(): PolyAsyncIterable<NonNullable<T>>;
    /**
     * Return an iteration of the result of calling `func(element)` for every element in `this`.
     *
     * @typeParam U - The return type of `func` and the generic type of the resulting iterable
     * @param func - A function that takes an element of `this` and returns something else
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} that yields the results of calling `func(element)`
     * for every element of `this`
     */
    map<U>(func: AsyncIndexedMapping<T, U>, options?: ConcurrencyOptions): PolyAsyncIterable<U>;
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the first element of the pair (the *key*) and preserving the second (the *value*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K1 - The key type of the original iterable, as passed to `func`
     * @typeParam K2 - The key type of the resulting iterable, as returned from `func`
     * @typeParam V - The value type of the original and resulting iterables, as passed to `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this` and using it to replace the keys
     */
    mapKeys<K1, K2, V>(this: PolyAsyncIterable<[K1, V]>, func: AsyncIndexedMapping<[K1, V], K2>, options?: ConcurrencyOptions): PolyAsyncIterable<[K2, V]>;
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the second element of the pair (the *value*) and preserving the first (the *key*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K - The key type of the original and resulting iterables, as passed to `func`
     * @typeParam V1 - The value type of the original iterable, as passed to `func`
     * @typeParam V2 - The value type of the resulting iterable, as returned from `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this` and using it to replace the values
     */
    mapValues<K, V1, V2>(this: PolyAsyncIterable<[K, V1]>, func: AsyncIndexedMapping<[K, V1], V2>, options?: ConcurrencyOptions): PolyAsyncIterable<[K, V2]>;
    /**
     * Return an iteration of the same elements as `this` after calling `func(element)` for all elements.
     *
     * @typeParam U - The return type of `func`
     * @param func - A function called for all elements
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} that yields the same elements as `this`
     */
    tap(func: AsyncIndexedRunnable<T>, options?: ConcurrencyOptions): PolyAsyncIterable<T>;
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolyAsyncIterable} that will yield the elements of all sub-iterables
     */
    flatten<U>(this: PolyAsyncIterable<Iterable<U> | AsyncIterable<U>>): PolyAsyncIterable<U>;
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @remarks
     * This method is an alias of {@link PolyAsyncIterable.flatten}.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolyAsyncIterable} that will yield the elements of all sub-iterables
     */
    flat<U>(this: PolyAsyncIterable<Iterable<U> | AsyncIterable<U>>): PolyAsyncIterable<U>;
    /**
     * Return an iteration of elements of the sub-iterables that result from calling `func(element)`
     * for every element in `this`.
     *
     * @remarks
     * This method is equivalent to calling {@link PolyAsyncIterable.map | map(func)}
     * and then {@link PolyAsyncIterable.flatten | flatten()}
     *
     * @typeParam U - The type of the sub-iterables returned by `func`
     * @param func - A function that takes an element of `this` and returns an iterable
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} that yields the elements of the subiterables that results from
     * calling `func(element)` for every element of `this`
     */
    flatMap<U>(func: AsyncIndexedMapping<T, Iterable<U> | AsyncIterable<U>>, options?: ConcurrencyOptions): PolyAsyncIterable<U>;
    /**
     * Return an iteration of arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this` iteration.
     *
     * @remarks
     * All chunks except possibly the last one will have exactly `num` elements.  The last chunk will have less elements
     * if the number of elements in the iteration is not divisible by `num`.  No chunk will ever be returned empty or
     * have more than `num` elements.
     *
     * @param num - Size of the chunks
     * @returns A new {@link PolyAsyncIterable} that yields arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this`
     */
    chunk(num: number): PolyAsyncIterable<Array<T>>;
    /**
     * Return an iteration of arrays with elements of this separated based on the result of calling `func(elements)`.
     *
     * @remarks
     * The chunking process works by keeping an open _current chunk_ and calling `func` to decide whether the next
     * element of the iteration will be part of the _current chunk_ or if it will be part of a new chunk.
     *
     * To do this, `func` will be called for every element of the iteration except the first, which will automatically
     * become part of the first chunk.  If `func` returns `true`, the element will be part of the current chunk, and
     * if it returns `false`, the _current chunk_ is closed and the element becomes the first element if the new
     * _current chunk_.
     * The arguments passed to `func` will be, in order:
     *   - `elem` - The element being currently processed
     *   - `lastElem` - The last element that was added to the current chunk
     *   - `firstElem` - The first element of the current chunk (might be the same as `lastElem`)
     *
     * All elements will be part of a chunk, and no chunk will ever be empty.
     *
     * @param func - A function that decides if an element is part of the current chunk or initiates a new one
     * @returns A new {@link PolyAsyncIterable} that yields arrays with the elements of `this` as separated by `func`
     */
    chunkWhile(func: AsyncChunkingPredicate<T>): PolyAsyncIterable<Array<T>>;
    /**
     * Return an iteration of group pairs, where the first element is a _group key_ and the second is an iterable of all
     * the elements for which `func(element)` returned the key.
     *
     * @remarks
     * This method is intended to be combined with {@link PolyAsyncIterable.toObject | toObject}
     * or {@link PolyAsyncIterable.toMap | toMap}, thus behaving like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupBy | Array.groupBy} and
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupByToMap | Array.groupByToMap}
     * respectively, but without losing the ablity to further process the iteration, such as by mapping, filtering, etc.
     *
     * @typeParam K - Type of the keys used to group elements
     * @param func - A function that returns the grouping key of each element
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} of group pairs with the key and the group
     */
    groupBy<K>(func: AsyncIndexedMapping<T, K>, options?: ConcurrencyOptions): PolyAsyncIterable<[K, Array<T>]>;
    /**
     * Return an iteration of unique elements, where two elements are considered equal if the result of `func(element)` is
     * the same for both elements.
     *
     * @remarks
     * Note that the first element seen with a specific key is always the one yielded, and every other element afterwards
     * is ignored.
     *
     * If no key-mapping function is given, the elements theselves are used as keys.
     * This is likely _not_ what you want in most situations unless elements are primitive types.
     *
     * @param func - A function that returns a _key_ used for uniqueness checks.
     * If not passed, an identitity function is used.
     * @param options - Options for concurrency of this operation
     * @returns A new {@link PolyAsyncIterable} only elements for which `func(element)` returns a value that hasn't
     * been seen before
     */
    unique(func?: AsyncIndexedMapping<T, unknown>, options?: ConcurrencyOptions): PolyAsyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` in reverse order.
     *
     * @remarks
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @returns A new {@link PolyAsyncIterable} that yields the elements of `this` in reverse order
     */
    reverse(): PolyAsyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` sorted according to `func`
     *
     * @remarks
     * The sort function `func` is used to call
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * on an array of all the elements.
     * However, the default comparator function will sort elements according to the `<` and `>` operators defined on
     * their own type, of always sorting lexicagraphically like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * does.
     *
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @param func - A comparator function
     * @returns A new {@link PolyAsyncIterable} that yields the elements of `this` sorted according to `func`
     */
    sort(func?: Comparator<T>): PolyAsyncIterable<T>;
    /**
     * Return an array of all elements of this iteration in the same order that were yielded.
     *
     * @returns A promise to an array that contains the same elements as this iteration, in the same order
     */
    toArray(): Promise<Array<T>>;
    /**
     * Splits this iteration into two arrays, one with elements for which `func(element)` returns `true` (the _truthy
     * elements_) and one for which it returns `false` (the _falsy elements_).
     *
     * @remarks
     * The array of _truthy elements_ has its element type narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A function that will be called for all elements to split them into the result arrays
     * @param options - Options for concurrency of this operation
     * @returns A promise to a tuple with the array of values for which `func` returned `true` as the first element, and
     * the array of values for which `func` returned `false` as the second element.
     */
    toPartitionArrays<U extends T>(func: IndexedTypePredicate<T, U>, options?: ConcurrencyOptions): Promise<[Array<U>, Array<Exclude<T, U>>]>;
    /**
     * Splits this iteration into two arrays, one with elements for which `func(element)` returns `true` (the _truthy
     * elements_) and one for which it returns `false` (the _falsy elements_).
     *
     * @param func - A function that will be called for all elements to split them into the result arrays
     * @param options - Options for concurrency of this operation
     * @returns A promise to a tuple with the array of values for which `func` returned `true` as the first element, and
     * the array of values for which `func` returned `false` as the second element.
     */
    toPartitionArrays(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<[Array<T>, Array<T>]>;
    /**
     * Return an object made from the entries of `this`.
     * This method is roughly equivalent to calling `Object.fromEntires(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs where the first component is a valid object key type.
     *
     * @returns A promise to an object composed of the entries yielded by this iterable.
     */
    toObject<K extends PropertyKey, V>(this: PolyAsyncIterable<[K, V] | readonly [K, V]>): Promise<Record<K, V>>;
    /**
     * Return a `Map` made from the entries of `this`.
     * This method is roughly equivalent to calling `new Map(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs where the first component is a valid object key type.
     *
     * @returns A promise to a `Map` composed of the entries yielded by this iterable.
     */
    toMap<K, V>(this: PolyAsyncIterable<[K, V] | readonly [K, V]>): Promise<Map<K, V>>;
    /**
     * Returns the first element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * The return type of this function is narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A type predicate called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise to the first element of the iteration for which `func` returned `true`
     */
    find<U extends T>(func: IndexedTypePredicate<T, U>, options?: ConcurrencyOptions): Promise<U | undefined>;
    /**
     * Returns the first element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise to the first element of the iteration for which `func` returned `true`
     */
    find(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<T | undefined>;
    /**
     * Returns the last element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * The return type of this function is narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A type predicate called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise to the last element of the iteration for which `func` returned `true`
     */
    findLast<U extends T>(func: IndexedTypePredicate<T, U>, options?: ConcurrencyOptions): Promise<U | undefined>;
    /**
     * Returns the last element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise to the last element of the iteration for which `func` returned `true`
     */
    findLast(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<T | undefined>;
    /**
     * Returns the index of the first element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise the index of the first element of the iteration for which `func` returned `true`
     */
    findIndex(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<number>;
    /**
     * Returns the index of the last element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @param options - Options for concurrency of this operation
     * @returns A promise the index of the last element of the iteration for which `func` returned `true`
     */
    findLastIndex(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<number>;
    /**
     * Returns whether an element is present in this iteration.
     *
     * @remarks
     * If the element is found in the iteration, no more elements are iterated.
     *
     * @param obj - The element to search in the iteration
     * @returns A promise to whether `obj` is present in this iteration or not
     */
    includes(obj: T): Promise<boolean>;
    /**
     * Returns `true` if calling `func(element)` returns `true` for at least one element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `true`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @param options - Options for concurrency of this operation
     * @returns A promise to whether calling `func` returned `true` on at least one element.
     */
    some(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<boolean>;
    /**
     * Returns `true` if calling `func(element)` returns `true` for every element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `false`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @param options - Options for concurrency of this operation
     * @returns A promise to whether calling `func` returned `true` for all elements.
     */
    every(func: AsyncIndexedPredicate<T>, options?: ConcurrencyOptions): Promise<boolean>;
    /**
     * Returns the result of calling the passed `reducer` for all elements of the iteration and the result of the
     * previous call to `reducer`, starting by passing `init` or, if not present, the first element of the iteration.
     *
     * @remarks
     * If the `init` argument is not present, at least one element must be present in the iteration, else an error will
     * be thrown
     *
     * `reducer` will be called with the accumulated result, the next element of the iteration, and the index of the
     * iteration.  The resolved return value will be the value passed to the next call as the first argument, or the
     * value returned if no more elements remain.
     *
     * @param reducer - A function to call for all elements with the result of a previous call
     * @param init - First element to be passed to the `reducer` function
     * @returns A promise the result to continually call `reducer` with all elements and the previous result
     */
    reduce(reducer: AsyncIndexedReducer<T, T>, init?: T): Promise<T>;
    /**
     * Returns the result of calling the passed `reducer` for all elements of the iteration and the result of the
     * previous call to `reducer`, starting by passing `init`.
     *
     * @remarks
     * `reducer` will be called with the accumulated result, the next element of the iteration, and the index of the
     * iteration.  The resolved return value will be the value passed to the next call as the first argument, or the
     * value returned if no more elements remain.
     *
     * @param reducer - A function to call for all elements with the result of a previous call
     * @param init - First element to be passed to the `reducer` function
     * @returns A promise the result to continually call `reducer` with all elements and the previous result
     */
    reduce<U>(reducer: AsyncIndexedReducer<T, U>, init: U): Promise<U>;
    /**
     * Return the number of elements on this iteration.
     *
     * @returns A promise to the number of elements in this iteration
     */
    count(): Promise<number>;
    /**
     * Call a function for each element of `this` iteration.
     *
     * @param func - A function to be called for every element of the iteration
     * @param options - Options for concurrency of this operation
     * @returns A promise that will resolve when all calls have resolved
     */
    forEach(func: AsyncIndexedRunnable<T>, options?: ConcurrencyOptions): Promise<void>;
    /**
     * Return the result of joining the elements of `this` with the given `glue`, or `','` if no glue is given.
     *
     * @remarks
     * `null` or `undefined` elements are treated as empty strings.
     *
     * @param glue - The string to use for joining the elements
     * @param options - Options for concurrency of this operation
     * @returns A promise to a string concatenating all elements of `this` using the given `glue`
     */
    join(glue?: string, options?: ConcurrencyOptions): Promise<string>;
    /**
     * Perform this iteration doing nothing.
     *
     * @param options - Options for concurrency of this operation
     * @returns a promise that will resolve when the iteration is done
     */
    complete(options?: ConcurrencyOptions): Promise<void>;
    /**
     * Returns a tuple containing `num` iterables that will yield independent
     * copies of the elements yielded by `this`.
     *
     * @remarks
     * Note that, as with every other method of this class, this instance is unusable after calling this method.
     *
     * In order to provide a truly independent iteration for all returned iterables, a buffer is kept, which can grow as
     * big as the whole iteration in certain circumstances.  The buffer is filled as fast as the fastest iterable requests
     * new items, and emptied as fast as the slowest iterable consumes those items.
     *
     * @param num - the number of copies to be returned
     * @returns An array of `num` elements containing independent copies of this iterable
     */
    duplicate<N extends number>(num: N): Tuple<PolyAsyncIterable<T>, N>;
}

/**
 * A `SyncIterable<T>` with a suite of methods for transforming the iteration into other iterations or to get a single
 * result from it.
 *
 * The methods of this class are intended to resemble those of `Array`, with added utilities where appropriate and made
 * for any kind of iterable.
 *
 * @public
 */
export declare abstract class PolySyncIterable<T> implements Iterable<T> {
    /**
     * Allows this class to work as a regular `Iterable<T>`
     *
     * @returns an iterable that will yield the same elements as the iterable used to create this instance
     */
    abstract [Symbol.iterator](): Iterator<T>;
    /**
     * Return an async version of this same iteration.
     *
     * @returns A {@link PolyAsyncIterable} that yields the same elements as `this`
     */
    async(): PolyAsyncIterable<T>;
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolySyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    append<U>(other: Iterable<U>): PolySyncIterable<T | U>;
    /**
     * Return a new iteration that will iterate over `this`, then over `other`.
     *
     * @remarks
     * This method is an alias for {@link PolySyncIterable.append}.
     *
     * @typeParam U - Type of the elements to be appended
     * @param other - Iterable to be appended
     * @returns a new {@link PolySyncIterable} that yields the elements of `this` and then the elements of `other`
     */
    concat<U>(other: Iterable<U>): PolySyncIterable<T | U>;
    /**
     * Return a new iteration that will iterate over `other`, then over `this`.
     *
     * @remarks
     * The resulting iteration is of the combined generic type of `this` and `other`, allowing this method to merge the
     * types of two distinct iterables.
     *
     * @typeParam U - Type of the elements to be prepended
     * @param other - Iterable to be prepended
     * @returns a new {@link PolySyncIterable} that yields the elements of `other` and then the elements of `this`
     */
    prepend<U>(other: Iterable<U>): PolySyncIterable<T | U>;
    /**
     * Return a new iteration that skips the first `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, except for the first
     * `num` elements
     */
    drop(num?: number): PolySyncIterable<T>;
    /**
     * Return a new iteration that iterates only over the first `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolySyncIterable} that yields the first `num` elements elements of `this`
     */
    take(num?: number): PolySyncIterable<T>;
    /**
     * Return a new iteration that skips the last `num` elements.  If there were less than `num` elements in the
     * iteration, no elements are yielded.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to skip those if the iteration ends,
     * and so elements effectively get delayed by `num` elements.
     *
     * @param num - The number of elements to skip
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, except for the last
     * `num` elements
     */
    dropLast(num?: number): PolySyncIterable<T>;
    /**
     * Return a new iteration that iterates only over the last `num` elements.  If there were less than than `num`
     * elements in the iteration, all elements are yielded with no additions.
     *
     * @remarks
     * The returned iteration keeps a buffer of `num` elements internally in order to know which elements to keep.
     * and so elements effectively get delayed until the iteration ends.
     *
     * @param num - The number of elements to yield
     * @returns a new {@link PolySyncIterable} that yields the last `num` elements elements of `this`
     */
    takeLast(num?: number): PolySyncIterable<T>;
    /**
     * Return a new iteration that skips the first few elements for which `func(element)` returns `true`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this`, excepts the first few for
     * which`func(element)` returns `true`
     */
    dropWhile(func: IndexedPredicate<T>): PolySyncIterable<T>;
    /**
     * Return a new iteration that yields the first few elements for which `func(element)` returns `true`.
     *
     * @remarks
     * Because the `func` argument is a type predicate, the result iteration will have the type asserted by `func`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this` as long as `func(element)`
     * returns `true`, correctly narrowed to the type asserted by `func`
     */
    takeWhile<U extends T>(func: IndexedTypePredicate<T, U>): PolySyncIterable<U>;
    /**
     * Return a new iteration that yields the first few elements for which `func(element)` returns `true`.
     *
     * @param func - The function to call on the elements
     * @returns a new {@link PolySyncIterable} that yields the same the elements of `this` as long as `func(element)`
     * returns `true`
     */
    takeWhile(func: IndexedPredicate<T>): PolySyncIterable<T>;
    /**
     * Return a new iteration that starts from the `start`th element (included)
     * and ends at the `end`th element (excluded) of `this`.
     *
     * @remarks
     * Both `start` and `end` allow for negative values, in which case they refer to the nth-to-last element,
     * with n being the absolute value of the argument.  `end` might also be `undefined`, in which case the iteration is
     * not shortened on the end side, yielding up to the end, including the last element.
     * This mimics the behaviour of {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/slice | Array.slice}.
     *
     * This function will likely need a buffer, effectively delaying the yielding of elements for a while.
     *
     * @param start - The index of the first element returned
     * @param end - The index of the first element *not* returned, inclusive
     * @returns a new {@link PolySyncIterable} that yields the elements going that starts from the `start`th element
     * (included) and ends at the `end`th element (excluded) of `this`
     */
    slice(start: number, end?: number): PolySyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` for which `func(element)` returns `true`.
     *
     * @remarks
     * Because the `func` argument is a type predicate, the result iteration will have the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`, if any
     * @param func - The function to be called on all elements
     * @returns A new {@link PolySyncIterable} with only elements for which `func(element)` returned true, correctly
     * narrowed to the type asserted by `func`
     *
     * {@label FILTER_TYPEPRED}
     */
    filter<U extends T>(func: IndexedTypePredicate<T, U>): PolySyncIterable<U>;
    /**
     * Return an iteration of the elements of `this` for which `func(element)` returns `true`.
     *
     * @param func - The function to be called on all elements
     * @returns A new {@link PolySyncIterable} with only elements for which `func(element)` returned true
     */
    filter(func: IndexedPredicate<T>): PolySyncIterable<T>;
    /**
     * Return an iteration of all the elements as `this` that aren't `null` or `undefined`.
     *
     * @remarks
     * This function is a shortcut to calling {@link PolySyncIterable.filter.(:FILTER_TYPEPRED)} with a type predicate
     * function that correctly filters out `null` and `undefined` values from the iteration.  Note that other falsy values
     * will remain in the iteration, and that the return value is correctly typed to exclude `null` and `undefined`.
     *
     * @returns A new {@link PolySyncIterable} that yields the same elements as `this`
     * except for `null` or `undefined` values
     */
    filterNotNullish(): PolySyncIterable<NonNullable<T>>;
    /**
     * Return an iteration of the result of calling `func(element)` for every element in `this`.
     *
     * @typeParam U - The return type of `func` and the generic type of the resulting iterable
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this`
     */
    map<U>(func: IndexedMapping<T, U>): PolySyncIterable<U>;
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the first element of the pair (the *key*) and preserving the second (the *value*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K1 - The key type of the original iterable, as passed to `func`
     * @typeParam K2 - The key type of the resulting iterable, as returned from `func`
     * @typeParam V - The value type of the original and resulting iterables, as passed to `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)` for every element of
     * `this` and uses it to replace the keys
     */
    mapKeys<K1, K2, V>(this: PolySyncIterable<[K1, V]>, func: IndexedMapping<[K1, V], K2>): PolySyncIterable<[K2, V]>;
    /**
     * Return an iteration of the pairs resulting of calling `func(element)` for every element in `this` and using it as
     * the second element of the pair (the *value*) and preserving the first (the *key*).
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @typeParam K - The key type of the original and resulting iterables, as passed to `func`
     * @typeParam V1 - The value type of the original iterable, as passed to `func`
     * @typeParam V2 - The value type of the resulting iterable, as returned from `func`
     * @param func - A function that takes an element of `this` and returns something else
     * @returns A new {@link PolySyncIterable} that yields the results of calling `func(element)`
     * for every element of `this` and using it to replace the values
     */
    mapValues<K, V1, V2>(this: PolySyncIterable<[K, V1]>, func: IndexedMapping<[K, V1], V2>): PolySyncIterable<[K, V2]>;
    /**
     * Return an iteration of the same elements as `this` after calling `func(element)` for all elements.
     *
     * @typeParam U - The return type of `func`
     * @param func - A function called for all elements
     * @returns A new {@link PolySyncIterable} that yields the same elements as `this`
     */
    tap(func: IndexedRunnable<T>): PolySyncIterable<T>;
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolySyncIterable} that will yield the elements of all sub-iterables
     */
    flatten<U>(this: PolySyncIterable<Iterable<U>>): PolySyncIterable<U>;
    /**
     * Return an iteration of the yielded elements of the sub-iterables.
     *
     * @remarks
     * This method is an alias of {@link PolySyncIterable.flatten}.
     *
     * @typeParam U - The type of the sub-iterable elements
     * @returns A new {@link PolySyncIterable} that will yield the elements of all sub-iterables
     */
    flat<U>(this: PolySyncIterable<Iterable<U>>): PolySyncIterable<U>;
    /**
     * Return an iteration of elements of the sub-iterables that result from calling `func(element)`
     * for every element in `this`.
     *
     * @remarks
     * This method is equivalent to calling {@link PolySyncIterable.map | map(func)}
     * and then {@link PolySyncIterable.flatten | flatten()}
     *
     * @typeParam U - The type of the sub-iterables returned by `func`
     * @param func - A function that takes an element of `this` and returns an iterable
     * @returns A new {@link PolySyncIterable} that yields the elements of the subiterables that results from
     * calling `func(element)` for every element of `this`
     */
    flatMap<U>(func: IndexedMapping<T, Iterable<U>>): PolySyncIterable<U>;
    /**
     * Return an iteration of arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this` iteration.
     *
     * @remarks
     * All chunks except possibly the last one will have exactly `num` elements.  The last chunk will have less elements
     * if the number of elements in the iteration is not divisible by `num`.  No chunk will ever be returned empty or
     * have more than `num` elements.
     *
     * @param num - Size of the chunks
     * @returns A new {@link PolySyncIterable} that yields arrays of size `num` (except possibly the last) containing
     * groupings of elements of `this`
     */
    chunk(num?: number): PolySyncIterable<Array<T>>;
    /**
     * Return an iteration of arrays with elements of this separated based on the result of calling `func(elements)`.
     *
     * @remarks
     * The chunking process works by keeping an open _current chunk_ and calling `func` to decide whether the next
     * element of the iteration will be part of the _current chunk_ or if it will be part of a new chunk.
     *
     * To do this, `func` will be called for every element of the iteration except the first, which will automatically
     * become part of the first chunk.  If `func` returns `true`, the element will be part of the current chunk, and
     * if it returns `false`, the _current chunk_ is closed and the element becomes the first element if the new
     * _current chunk_.
     * The arguments passed to `func` will be, in order:
     *   - `elem` - The element being currently processed
     *   - `lastElem` - The last element that was added to the current chunk
     *   - `firstElem` - The first element of the current chunk (might be the same as `lastElem`)
     *
     * All elements will be part of a chunk, and no chunk will ever be empty.
     *
     * @param func - A function that decides if an element is part of the current chunk or initiates a new one
     * @returns A new {@link PolySyncIterable} that yields arrays with the elements of `this` as separated by `func`
     */
    chunkWhile(func: ChunkingPredicate<T>): PolySyncIterable<Array<T>>;
    /**
     * Return an iteration of group pairs, where the first element is a _group key_ and the second is an iterable of all
     * the elements for which `func(element)` returned the key.
     *
     * @remarks
     * This method is intended to be combined with {@link PolySyncIterable.toObject | toObject}
     * or {@link PolySyncIterable.toMap | toMap}, thus behaving like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupBy | Array.groupBy} and
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/groupByToMap | Array.groupByToMap}
     * respectively, but without losing the ablity to further process the iteration, such as by mapping, filtering, etc.
     *
     * @typeParam K - Type of the keys used to group elements
     * @param func - A function that returns the grouping key of each element
     * @returns A new {@link PolySyncIterable} of group pairs with the key and the group
     */
    groupBy<K>(func: IndexedMapping<T, K>): PolySyncIterable<[K, Array<T>]>;
    /**
     * Return an iteration of unique elements, where two elements are considered equal if the result of `func(element)` is
     * the same for both elements.
     *
     * @remarks
     * Note that the first element seen with a specific key is always the one yielded, and every other element afterwards
     * is ignored.
     *
     * If no key-mapping function is given, the elements theselves are used as keys.
     * This is likely _not_ what you want in most situations unless elements are primitive types.
     *
     * @param func - A function that returns a _key_ used for uniqueness checks.
     * If not passed, an identitity function is used.
     * @returns A new {@link PolySyncIterable} only elements for which `func(element)` returns a value that hasn't
     * been seen before
     */
    unique(func?: IndexedMapping<T, unknown>): PolySyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` in reverse order.
     *
     * @remarks
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @returns A new {@link PolySyncIterable} that yields the elements of `this` in reverse order
     */
    reverse(): PolySyncIterable<T>;
    /**
     * Return an iteration of the elements of `this` sorted according to `func`
     *
     * @remarks
     * The sort function `func` is used to call
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * on an array of all the elements.
     * However, the default comparator function will sort elements according to the `<` and `>` operators defined on
     * their own type, of always sorting lexicagraphically like
     * {@link https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array/sort | Array.sort}
     * does.
     *
     * This method will buffer _all_ elements of the iteration, and yield them all at once at the end
     *
     * @param func - A comparator function
     * @returns A new {@link PolySyncIterable} that yields the elements of `this` sorted according to `func`
     */
    sort(func?: Comparator<T>): PolySyncIterable<T>;
    /**
     * Return an array of all elements of this iteration in the same order that were yielded.
     *
     * @returns An array that contains the same elements as this iteration, in the same order
     */
    toArray(): Array<T>;
    /**
     * Splits this iteration into two arrays, one with elements for which `func(element)` returns `true` (the _truthy
     * elements_) and one for which it returns `false` (the _falsy elements_).
     *
     * @remarks
     * The array of _truthy elements_ has its element type narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A function that will be called for all elements to split them into the result arrays
     * @returns A tuple with the array of values for which `func` returned `true` as the first element, and the array of
     * values for which `func` returned `false` as the second element.
     */
    toPartitionArrays<U extends T>(func: IndexedTypePredicate<T, U>): [Array<U>, Array<Exclude<T, U>>];
    /**
     * Splits this iteration into two arrays, one with elements for which `func(element)` returns `true` (the _truthy
     * elements_) and one for which it returns `false` (the _falsy elements_).
     *
     * @param func - A function that will be called for all elements to split them into the result arrays
     * @returns A tuple with the array of values for which `func` returned `true` as the first element, and the array of
     * values for which `func` returned `false` as the second element.
     */
    toPartitionArrays(func: IndexedPredicate<T>): [Array<T>, Array<T>];
    /**
     * Return an object made from the entries of `this`.
     * This method is roughly equivalent to calling `Object.fromEntires(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs where the first component is a valid object key type.
     *
     * @returns An object composed of the entries yielded by this iterable.
     */
    toObject<K extends PropertyKey, V>(this: PolySyncIterable<[K, V] | readonly [K, V]>): Record<K, V>;
    /**
     * Return a `Map` made from the entries of `this`.
     * This method is roughly equivalent to calling `new Map(iter.toArray())`.
     *
     * @remarks
     * This method is only available for iterations of pairs.
     *
     * @returns A `Map` composed of the entries yielded by this iterable.
     */
    toMap<K, V>(this: PolySyncIterable<[K, V] | readonly [K, V]>): Map<K, V>;
    /**
     * Returns the first element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * The return type of this function is narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A type predicate called for elements of `this`
     * @returns The first element of the iteration for which `func` returned `true`
     */
    find<U extends T>(func: IndexedTypePredicate<T, U>): U | undefined;
    /**
     * Returns the first element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The first element of the iteration for which `func` returned `true`
     */
    find(func: IndexedPredicate<T>): T | undefined;
    /**
     * Returns the last element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * The return type of this function is narrowed to the type asserted by `func`.
     *
     * @typeParam U - The type asserted by `func`
     * @param func - A type predicate called for elements of `this`
     * @returns The last element of the iteration for which `func` returned `true`
     */
    findLast<U extends T>(func: IndexedTypePredicate<T, U>): U | undefined;
    /**
     * Returns the last element for which `func(element)` returns `true`, or `undefined` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The last element of the iteration for which `func` returned `true`
     */
    findLast(func: IndexedPredicate<T>): T | undefined;
    /**
     * Returns the index of the first element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on elements of this iteration until it returns `true`, and then not called again.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The index of the first element of the iteration for which `func` returned `true`
     */
    findIndex(func: IndexedPredicate<T>): number;
    /**
     * Returns the index of the last element for which `func(element)` returns `true`, or `-1` if it never does.
     *
     * @remarks
     * `func` will be called on *all* of this iteration, and the result will not be returned until the iteration ends.
     *
     * Note that this method is rather useless given that iterables are single-use and have no indexing capabilities,
     * but it's here for completion and consistency with `Array`.
     *
     * @param func - A boolean returning function called for elements of `this`
     * @returns The index of the last element of the iteration for which `func` returned `true`
     */
    findLastIndex(func: IndexedPredicate<T>): number;
    /**
     * Returns whether an element is present in this iteration.
     *
     * @remarks
     * If the element is found in the iteration, no more elements are iterated.
     *
     * @param obj - The element to search in the iteration
     * @returns Whether `obj` is present in this iteration or not
     */
    includes(obj: T): boolean;
    /**
     * Returns `true` if calling `func(element)` returns `true` for at least one element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `true`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @returns Whether calling `func` returned `true` on at least one element.
     */
    some(func: IndexedPredicate<T>): boolean;
    /**
     * Returns `true` if calling `func(element)` returns `true` for every element, and `false` otherwise
     *
     * @remarks
     * If a call to `func(element)` returns `false`, no more elements are iterated.
     *
     * @param func - A function to be called on the elements of the iteration
     * @returns Whether calling `func` returned `true` for all elements.
     */
    every(func: IndexedPredicate<T>): boolean;
    /**
     * Returns the result of calling the passed `reducer` for all elements of the iteration and the result of the
     * previous call to `reducer`, starting by passing `init` or, if not present, the first element of the iteration.
     *
     * @remarks
     * If the `init` argument is not present, at least one element must be present in the iteration, else an error will
     * be thrown
     *
     * `reducer` will be called with the accumulated result, the next element of the iteration, and the index of the
     * iteration.  The resolved return value will be the value passed to the next call as the first argument, or the
     * value returned if no more elements remain.
     *
     * @param reducer - A function to call for all elements with the result of a previous call
     * @param init - First element to be passed to the `reducer` function
     * @returns The result to continually call `reducer` with all elements and the previous result
     */
    reduce(reducer: IndexedReducer<T, T>, init?: T): T;
    /**
     * Returns the result of calling the passed `reducer` for all elements of the iteration and the result of the
     * previous call to `reducer`, starting by passing `init`.
     *
     * @remarks
     * `reducer` will be called with the accumulated result, the next element of the iteration, and the index of the
     * iteration.  The resolved return value will be the value passed to the next call as the first argument, or the
     * value returned if no more elements remain.
     *
     * @param reducer - A function to call for all elements with the result of a previous call
     * @param init - First element to be passed to the `reducer` function
     * @returns The result to continually call `reducer` with all elements and the previous result
     */
    reduce<U>(reducer: IndexedReducer<T, U>, init: U): U;
    /**
     * Return the number of elements on this iteration.
     *
     * @returns The number of elements in this iteration
     */
    count(): number;
    /**
     * Call a function for each element of `this` iteration.
     *
     * @param func - A function to be called for every element of the iteration
     */
    forEach(func: IndexedRunnable<T>): void;
    /**
     * Return the result of joining the elements of `this` with the given `glue`, or `','` if no glue is given.
     *
     * @remarks
     * `null` or `undefined` elements are treated as empty strings.
     *
     * @param glue - The string to use for joining the elements
     * @returns A string concatenating all elements of `this` using the given `glue`
     */
    join(glue?: string): string;
    /**
     * Perform this iteration doing nothing.
     */
    complete(): void;
    /**
     * Returns a tuple containing `num` iterables that will yield independent
     * copies of the elements yielded by `this`.
     *
     * @remarks
     * Note that, as with every other method of this class, this instance is unusable after calling this method.
     *
     * In order to provide a truly independent iteration for all returned iterables, a buffer is kept, which can grow as
     * big as the whole iteration in certain circumstances.  The buffer is filled as fast as the fastest iterable requests
     * new items, and emptied as fast as the slowest iterable requests new items.
     *
     * Note that for synchronous iterations, it's common to end up with a full buffer if the returned duplicated elements
     * are used in sequence.  In this situation, it might be more useful to simply convert the iteration to an array and
     * pass it around, rather than pay the overhead of this method.
     *
     * @param num - the number of copies to be returned
     * @returns An array of `num` elements containing independent copies of this iterable
     */
    duplicate<N extends number>(num: N): Tuple<PolySyncIterable<T>, N>;
}

/**
 * A tuple of `N` elements of type `T`.
 *
 * @typeParam T - The type of elements of the tuple
 * @typeParam N - The number of elements in this tuple
 */
export declare type Tuple<T, N extends number> = [T, ...T[]] & {
    length: N;
};

export { }
